

table{
    width: 100%;
    border-collapse: collapse;
    border-spacing: 0;
    align: center;
}

th:first-child{
    border-radius: 5px 0 0 0;
}

th:last-child{
    border-radius: 0 5px 0 0;
    border-right: 1px solid #3c6690;
}

th{
    text-align: center;
    color:white;
    background: linear-gradient(#829ebc,#225588);
    border-left: 1px solid #3c6690;
    border-top: 1px solid #3c6690;
    border-bottom: 1px solid #3c6690;
    box-shadow: 0px 1px 1px rgba(255,255,255,0.3) inset;
    width: 20%;
    padding: 10px 0;
}

td{
    padding: 10px 0;
    text-align: center;
}

tr:nth-child(odd){
    background-color: #cccccc;
}
